/**********************************************************************
Copyright (c) 2009 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.datanucleus.store.odf;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.FetchPlan;
import org.datanucleus.NucleusContext;
import org.datanucleus.exceptions.NucleusDataStoreException;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.exceptions.NucleusUserException;
import org.datanucleus.identity.IdentityUtils;
import org.datanucleus.identity.OID;
import org.datanucleus.identity.OIDFactory;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.ColumnMetaData;
import org.datanucleus.metadata.EmbeddedMetaData;
import org.datanucleus.metadata.IdentityMetaData;
import org.datanucleus.metadata.IdentityType;
import org.datanucleus.metadata.MetaData;
import org.datanucleus.metadata.MetaDataManager;
import org.datanucleus.metadata.Relation;
import org.datanucleus.metadata.VersionMetaData;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.FieldValues;
import org.datanucleus.store.ObjectProvider;
import org.datanucleus.store.connection.ManagedConnection;
import org.datanucleus.store.odf.fieldmanager.FetchFieldManager;
import org.datanucleus.util.Localiser;
import org.datanucleus.util.NucleusLogger;

import org.odftoolkit.odfdom.doc.OdfSpreadsheetDocument;
import org.odftoolkit.odfdom.incubator.doc.office.OdfOfficeAutomaticStyles;
import org.odftoolkit.odfdom.incubator.doc.style.OdfStyle;
import org.odftoolkit.odfdom.pkg.OdfFileDom;
import org.odftoolkit.odfdom.doc.table.OdfTable;
import org.odftoolkit.odfdom.doc.table.OdfTableCell;
import org.odftoolkit.odfdom.doc.table.OdfTableRow;
import org.odftoolkit.odfdom.dom.attribute.office.OfficeValueTypeAttribute;
import org.odftoolkit.odfdom.dom.style.OdfStyleFamily;

/**
 * Utilities to assist in persistence to ODF spreadsheets.
 */
public class ODFUtils
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER = Localiser.getInstance(
        "org.datanucleus.store.odf.Localisation", ODFStoreManager.class.getClassLoader());

    /**
     * Convenience method to return the required sheet name for the specified class.
     * @param cmd Metadata for the class
     * @return The name of the sheet (table)
     */
    public static String getSheetNameForClass(AbstractClassMetaData cmd)
    {
        String sheetName = cmd.getValueForExtension("sheet");
        if (sheetName == null)
        {
            if (cmd.getTable() != null)
            {
                sheetName = cmd.getTable();
            }
            else
            {
                sheetName = cmd.getName();
            }
        }
        return sheetName;
    }

    /**
     * Convenience method to find the sheet (table) node appropriate for the specified class.
     * @param cmd Metadata for the class
     * @param spreadsheetDoc The spreadsheet document
     * @return The table node (or null of not present)
     */
    public static OdfTable getODFTableForClass(AbstractClassMetaData cmd, OdfSpreadsheetDocument spreadsheetDoc)
    {
        String sheetNameForClass = getSheetNameForClass(cmd);
        OdfTable table = spreadsheetDoc.getTableByName(sheetNameForClass);
        return table;
    }

    /**
     * Convenience method to find the row of an object in the provided sheet.
     * For application-identity does a search for a row with the specified PK field values.
     * For datastore-identity does a search for the row with the datastore column having the specified value
     * @param sm ObjectProvider for the object
     * @param spreadsheetDoc The spreadsheet document
     * @return The row (or null if not found)
     */
    public static OdfTableRow getTableRowForObjectInSheet(ObjectProvider sm, OdfSpreadsheetDocument spreadsheetDoc)
    {
        final AbstractClassMetaData cmd = sm.getClassMetaData();
        OdfTable table = getODFTableForClass(sm.getClassMetaData(), spreadsheetDoc);
        if (table == null)
        {
            return null;
        }

        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            int[] pkFieldNumbers = cmd.getPKMemberPositions();
            Object[] pkFieldValues = new Object[pkFieldNumbers.length];
            for (int i=0;i<pkFieldNumbers.length;i++)
            {
                pkFieldValues[i] = sm.provideField(pkFieldNumbers[i]);
            }

            List<OdfTableRow> rows = table.getRowList();
            Iterator<OdfTableRow> rowIter = rows.iterator();
            while (rowIter.hasNext())
            {
                OdfTableRow row = rowIter.next();
                boolean isRow = true;

                for (int j=0;j<pkFieldNumbers.length;j++)
                {
                    int index = getColumnPositionForFieldOfClass(cmd, pkFieldNumbers[j]);
                    OdfTableCell cellNode = row.getCellByIndex(index);
                    Object pkVal = pkFieldValues[j];
                    if (!doesCellMatchValue(cellNode, pkVal))
                    {
                        isRow = false;
                        break;
                    }
                }
                if (isRow)
                {
                    return row;
                }
            }
        }
        else if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            OID oid = (OID)sm.getInternalObjectId();
            Object key = oid.getKeyValue();
            int index = getColumnPositionForFieldOfClass(cmd, -1);
            List<OdfTableRow> rows = table.getRowList();
            Iterator<OdfTableRow> rowIter = rows.iterator();
            while (rowIter.hasNext())
            {
                OdfTableRow row = rowIter.next();
                OdfTableCell cell = row.getCellByIndex(index);
                if (doesCellMatchValue(cell, key))
                {
                    return row;
                }
            }
        }

        return null;
    }

    /**
     * Convenience method to return if the supplied cell matches the passed value.
     * Compares the type and value.
     * @param cell The cell
     * @param value The value
     * @return Whether they match
     */
    private static boolean doesCellMatchValue(OdfTableCell cell, Object value)
    {
        if (cell == null)
        {
            return false;
        }

        if (value instanceof Long && isOfficeValueTypeConsistent(cell, OfficeValueTypeAttribute.Value.FLOAT) && 
            cell.getDoubleValue().longValue() == ((Long)value).longValue())
        {
            return true;
        }
        else if (value instanceof Integer && isOfficeValueTypeConsistent(cell, OfficeValueTypeAttribute.Value.FLOAT) &&
            cell.getDoubleValue().intValue() == ((Integer)value).intValue())
        {
            return true;
        }
        else if (value instanceof Short && isOfficeValueTypeConsistent(cell, OfficeValueTypeAttribute.Value.FLOAT) &&
            cell.getDoubleValue().shortValue() == ((Short)value).shortValue())
        {
            return true;
        }
        else if (value instanceof String && isOfficeValueTypeConsistent(cell, OfficeValueTypeAttribute.Value.STRING) && 
            cell.getStringValue().equals((String)value))
        {
            return true;
        }
        // TODO Cater for other types
        return false;
    }

    /**
     * Method to return the metadata for the member at the specified column position.
     * @param acmd Metadata for the class
     * @param colIndex Column position
     * @return Metadata (if datastore-id is DatastoreIdentityMetaData, if version is VersionMetaData, else
     *     is AbstractMemberMetaData)
     */
    public static MetaData getMetaDataForColumnIndex(AbstractClassMetaData acmd, int colIndex)
    {
        int numFields = acmd.getAllMemberPositions().length;
        for (int i=0;i<numFields;i++)
        {
            int colIndexForField = getColumnPositionForFieldOfClass(acmd, i);
            if (colIndexForField == colIndex)
            {
                return acmd.getMetaDataForManagedMemberAtAbsolutePosition(i);
            }
        }
        if (acmd.getIdentityType() == IdentityType.DATASTORE && getColumnPositionForFieldOfClass(acmd, -1) == colIndex)
        {
            return acmd.getIdentityMetaData();
        }
        if (acmd.isVersioned() && getColumnPositionForFieldOfClass(acmd, -2) == colIndex)
        {
            return acmd.getVersionMetaDataForClass();
        }
        return null;
    }

    /**
     * Convenience method to get the position where a field of a class is persisted.
     * Uses the column "position" attribute if defined, otherwise uses the column name (as an integer).
     * The field number is the absolute number (0 or higher); a value of -1 implies surrogate identity column,
     * and -2 implies surrogate version column
     * @param acmd MetaData for the class
     * @param inputFieldNumber Absolute field number that we are interested in (-1 = datastore-id, -2=version)
     */
    public static int getColumnPositionForFieldOfClass(AbstractClassMetaData acmd, int inputFieldNumber)
    {
        int fieldNumber = inputFieldNumber;
        if (inputFieldNumber == -1)
        {
            // Datastore-identity, so allocate next column after normal fields
            fieldNumber = acmd.getNoOfManagedMembers() + acmd.getNoOfInheritedManagedMembers();
        }
        else if (inputFieldNumber == -2)
        {
            // Version, so allocate next column after normal fields (and optionally datastore-identity)
            if (acmd.getIdentityType() == IdentityType.DATASTORE)
            {
                fieldNumber = acmd.getNoOfManagedMembers() + acmd.getNoOfInheritedManagedMembers() + 1;
            }
            else
            {
                fieldNumber = acmd.getNoOfManagedMembers() + acmd.getNoOfInheritedManagedMembers();
            }
        }

        if (inputFieldNumber >= 0)
        {
            // Field of the class
            AbstractMemberMetaData ammd = acmd.getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
            Integer colPos = (ammd.getColumnMetaData() == null || ammd.getColumnMetaData().length == 0 ? 
                    null : ammd.getColumnMetaData()[0].getPosition());
            if (colPos == null)
            {
                ColumnMetaData[] colmds = ammd.getColumnMetaData();
                if (colmds != null && colmds.length > 0)
                {
                    String colName = colmds[0].getName();
                    try
                    {
                        return new Integer(colName).intValue();
                    }
                    catch (NumberFormatException nfe)
                    {
                        return (int)fieldNumber;
                    }
                }
                else
                {
                    return (int)fieldNumber;
                }
            }
            else
            {
                return colPos.intValue();
            }
        }
        else if (inputFieldNumber == -1)
        {
            // Surrogate datastore identity column
            IdentityMetaData imd = acmd.getIdentityMetaData();
            if (imd != null)
            {
                Integer colPos = (imd.getColumnMetaData() == null ? null : imd.getColumnMetaData().getPosition());
                if (colPos == null)
                {
                    if (imd.getColumnMetaData() != null)
                    {
                        String colName = imd.getColumnMetaData().getName();
                        try
                        {
                            return new Integer(colName).intValue();
                        }
                        catch (NumberFormatException nfe)
                        {
                            return (int)fieldNumber;
                        }
                    }
                    else
                    {
                        return (int)fieldNumber;
                    }
                }
                else
                {
                    return colPos.intValue();
                }
            }
            return -1;
        }
        else if (inputFieldNumber == -2)
        {
            // Surrogate version column
            VersionMetaData vmd = acmd.getVersionMetaDataForClass();
            if (vmd != null)
            {
                Integer colPos = (vmd.getColumnMetaData() == null ? null : vmd.getColumnMetaData().getPosition());
                if (colPos == null)
                {
                    if (vmd.getColumnMetaData() != null)
                    {
                        String colName = vmd.getColumnMetaData().getName();
                        try
                        {
                            return new Integer(colName).intValue();
                        }
                        catch (NumberFormatException nfe)
                        {
                            return (int)fieldNumber;
                        }
                    }
                    else
                    {
                        return (int)fieldNumber;
                    }
                }
                else
                {
                    return colPos.intValue();
                }
            }
            return -1;
        }
        else
        {
            throw new NucleusException("Unsupported field number " + fieldNumber);
        }
    }

    /**
     * Return the column position for the specified field of an embedded object stored in the specified owner field.
     * Uses the column "position" attribute if defined, otherwise uses the column name (as an integer).
     * @param fieldNumber Number of the field in the embedded class.
     * @param ownerMmd the owner field
     * @return The column position
     */
    public static int getColumnPositionForFieldOfEmbeddedClass(int fieldNumber, 
            AbstractMemberMetaData ownerMmd)
    {
        if (fieldNumber >= 0)
        {
            EmbeddedMetaData emd = ownerMmd.getEmbeddedMetaData();
            AbstractMemberMetaData[] emb_mmd = emd.getMemberMetaData();
            AbstractMemberMetaData ammd = emb_mmd[fieldNumber];
            Integer colPos = (ammd.getColumnMetaData() == null || ammd.getColumnMetaData().length == 0 ?
                    null : ammd.getColumnMetaData()[0].getPosition());
            if (colPos == null)
            {
                ColumnMetaData[] colmds = ammd.getColumnMetaData();
                if (colmds != null && colmds.length > 0)
                {
                    String colName = colmds[0].getName();
                    try
                    {
                        return Integer.valueOf(colName).intValue();
                    }
                    catch (NumberFormatException nfe)
                    {
                        return (int)fieldNumber;
                    }
                }
                else
                {
                    return (int)fieldNumber;
                }
            }
            else
            {
                return colPos;
            }
        }
        else
        {
            throw new NucleusException("Unsupported field number " + fieldNumber + " of owner "+ ownerMmd.getFullFieldName());
        }
    }

    /**
     * Convenience method to get all objects of the candidate type (and optional subclasses) from the 
     * specified workbook connection.
     * @param ec execution context
     * @param mconn Managed Connection
     * @param candidateClass Candidate
     * @param subclasses Include subclasses?
     * @param ignoreCache Whether to ignore the cache
     * @return List of objects of the candidate type
     */
    public static List getObjectsOfCandidateType(ExecutionContext ec, ManagedConnection mconn, 
            Class candidateClass, boolean subclasses, boolean ignoreCache)
    {
        OdfSpreadsheetDocument spreadsheetDoc = (OdfSpreadsheetDocument)mconn.getConnection();
        List results = null;
        try
        {
            ClassLoaderResolver clr = ec.getClassLoaderResolver();
            final AbstractClassMetaData acmd = ec.getMetaDataManager().getMetaDataForClass(candidateClass, clr);
            results = getObjectsOfCandidateType(ec, spreadsheetDoc, acmd, ignoreCache);
            if (subclasses)
            {
                // Add on any subclass objects
                String[] subclassNames = ec.getMetaDataManager().getSubclassesForClass(candidateClass.getName(), true);
                if (subclassNames != null)
                {
                    for (int i=0;i<subclassNames.length;i++)
                    {
                        AbstractClassMetaData cmd =
                            ec.getMetaDataManager().getMetaDataForClass(subclassNames[i], clr);
                        results.addAll(getObjectsOfCandidateType(ec, spreadsheetDoc, cmd, ignoreCache));
                    }
                }
            }
        }
        catch (Exception e)
        {
            throw new NucleusDataStoreException("Error extracting results from ODF document", e);
        }
        return results;
    }

    /**
     * Method to extract the actual objects of a particular class from the Excel spreadsheet.
     * @param ec execution context
     * @param spreadsheetDoc Spreadsheet
     * @param acmd MetaData for the class
     * @param ignoreCache Whether to ignore the cache
     * @return List of objects (connected to StateManagers as required)
     */
    private static List getObjectsOfCandidateType(ExecutionContext ec, OdfSpreadsheetDocument spreadsheetDoc,
            final AbstractClassMetaData acmd, boolean ignoreCache)
    {
        List results = new ArrayList();

        final OdfTable table = ODFUtils.getODFTableForClass(acmd, spreadsheetDoc);
        if (table != null)
        {
            List<OdfTableRow> rows = table.getRowList();
            NucleusLogger.GENERAL.info(">> ODFUtils.getObjectsOfCandidate numRows=" + rows.size() + " table=" + getSheetNameForClass(acmd));
            Iterator<OdfTableRow> rowIter = rows.iterator();
            while (rowIter.hasNext())
            {
                final OdfTableRow row = rowIter.next();
                OdfStyle style = row.getDefaultCellStyle();
                NucleusLogger.GENERAL.info(">> row=" + row + " style=" + style);
                String styleName = (style != null ? style.getStyleNameAttribute() : null);
                if (styleName != null && styleName.equals("DN_Headers"))
                {
                    // Skip headers
                    continue;
                }

                if (acmd.getIdentityType() == IdentityType.APPLICATION)
                {
                    Object id = IdentityUtils.getApplicationIdentityForResultSetRow(ec, acmd, null, false, 
                        new FetchFieldManager(ec, acmd, row));
                    results.add(ec.findObject(id, new FieldValues()
                    {
                        // StateManager calls the fetchFields method
                        public void fetchFields(ObjectProvider sm)
                        {
                            sm.replaceFields(acmd.getAllMemberPositions(), 
                                new FetchFieldManager(sm, row));
                        }
                        public void fetchNonLoadedFields(ObjectProvider sm)
                        {
                            sm.replaceNonLoadedFields(acmd.getAllMemberPositions(), 
                                new FetchFieldManager(sm, row));
                        }
                        public FetchPlan getFetchPlanForLoading()
                        {
                            return null;
                        }
                    }, null, ignoreCache));
                }
                else if (acmd.getIdentityType() == IdentityType.DATASTORE)
                {
                    int idIndex = ODFUtils.getColumnPositionForFieldOfClass(acmd, -1);
                    OdfTableCell idCell = row.getCellByIndex(idIndex);
                    Object idKey = null;
                    if (isOfficeValueTypeConsistent(idCell, OfficeValueTypeAttribute.Value.STRING))
                    {
                        idKey = idCell.getStringValue();
                    }
                    else
                    {
                        idKey = new Long(idCell.getDoubleValue().longValue());
                    }
                    OID oid = OIDFactory.getInstance(ec.getNucleusContext(), acmd.getFullClassName(), idKey);
                    // TODO This ignores the fields in this row!
                    results.add(ec.findObject(oid, true, true, null));
                }
            }
        }
        return results;
    }

    public static boolean isOfficeValueTypeConsistent(OdfTableCell cell, OfficeValueTypeAttribute.Value type)
    {
        String cellTypeStr = cell.getValueType();
        String typeStr = type.toString();
        if (cellTypeStr != null && typeStr != null)
        {
            return cellTypeStr.equals(typeStr);
        }
        return false;
    }

    /**
     * Convenience method to add a table (worksheet) to represent a class.
     * @param doc The spreadsheet doc
     * @param cmd Metadata for the class
     * @param sheetName Name of the sheet
     * @return The table
     */
    public static OdfTable addTableForClass(OdfSpreadsheetDocument doc, AbstractClassMetaData cmd, String sheetName, 
            NucleusContext nucContext)
    {
        OdfFileDom contentDoc;
        try
        {
            contentDoc = doc.getContentDom();
        }
        catch (Exception e)
        {
            throw new NucleusDataStoreException("Exception thrown adding worksheet " + sheetName, e);
        }
        OdfOfficeAutomaticStyles styles = contentDoc.getAutomaticStyles();
        OdfStyle headerStyle = styles.getStyle("DN_Headers", OdfStyleFamily.TableCell);

        Map<Integer, String> colNameByPosition = new HashMap<Integer, String>();
        getColumnInformationForClass(colNameByPosition, cmd, nucContext.getClassLoaderResolver(null), 
            nucContext.getMetaDataManager());
        int numCols = colNameByPosition.size();

        OdfTable table = OdfTable.newTable(doc, 1, numCols);
        table.setTableName(sheetName);

        // Set the header row if required TODO Make this optional when ODFDOM allows tables with no rows/columns
        if (true)
        {
            OdfTableRow headerRow = table.getRowByIndex(0);
            headerRow.setDefaultCellStyle(headerStyle);
            Iterator<Map.Entry<Integer, String>> colIter = colNameByPosition.entrySet().iterator();
            while (colIter.hasNext())
            {
                Map.Entry<Integer, String> entry = colIter.next();
                int position = entry.getKey();
                if (position >= numCols)
                {
                    throw new NucleusUserException("Error in specification of column positions." +
                        " Class " + cmd.getFullClassName() + " has " + numCols + " in total" +
                        " yet there is a column defined at position=" + position);
                }
                String colName = entry.getValue();
                OdfTableCell cell = headerRow.getCellByIndex(position);
                cell.setValueType(OfficeValueTypeAttribute.Value.STRING.toString());
                cell.setStringValue(colName);
            }
        }

        if (NucleusLogger.DATASTORE_PERSIST.isDebugEnabled())
        {
            NucleusLogger.DATASTORE_PERSIST.debug(LOCALISER.msg("ODF.Insert.SheetCreated", sheetName));
        }

        return table;
    }

    protected static void getColumnInformationForClass(Map<Integer, String> colNameByPosition, 
            AbstractClassMetaData cmd, ClassLoaderResolver clr, MetaDataManager mmgr)
    {
        int numFields = cmd.getAllMemberPositions().length;
        for (int i=0;i<numFields;i++)
        {
            AbstractMemberMetaData mmd = cmd.getMetaDataForManagedMemberAtAbsolutePosition(i);
            int relationType = mmd.getRelationType(clr);
            if (Relation.isRelationSingleValued(relationType) && mmd.isEmbedded())
            {
                AbstractClassMetaData relCmd = mmgr.getMetaDataForClass(mmd.getType(), clr);
                getColumnInformationForEmbeddedClass(colNameByPosition, relCmd, mmd, clr, mmgr);
            }
            else
            {
                int index = ODFUtils.getColumnPositionForFieldOfClass(cmd, i);
                String name = mmd.getName();
                ColumnMetaData[] colmds = mmd.getColumnMetaData();
                if (colmds != null && colmds.length > 0 && colmds[0].getName() != null)
                {
                    name = colmds[0].getName();
                }
                String oldName = colNameByPosition.put(index, name);
                if (oldName != null)
                {
                    throw new NucleusUserException("Error assigning positions of columns in ODF." +
                        " Position=" + index + " was previously assigned to column with name=" + oldName +
                        " but now is for name=" + name +
                        " Check your metadata specification of positions");
                }
            }
        }
        if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            int index = ODFUtils.getColumnPositionForFieldOfClass(cmd, -1);
            String name = "ID";
            ColumnMetaData colmd = cmd.getIdentityMetaData().getColumnMetaData();
            if (colmd != null && colmd.getName() != null)
            {
                name = colmd.getName();
            }
            String oldName = colNameByPosition.put(index, name);
            if (oldName != null)
            {
                throw new NucleusUserException("Error assigning positions of columns in ODF." +
                    " Position=" + index + " was previously assigned to column with name=" + oldName +
                    " but now is for name=" + name +
                    " Check your metadata specification of positions");
            }
        }
        if (cmd.isVersioned())
        {
            int index = ODFUtils.getColumnPositionForFieldOfClass(cmd, -2);
            String name = "VERSION";
            VersionMetaData vermd = cmd.getVersionMetaDataForClass();
            ColumnMetaData colmd = vermd.getColumnMetaData();
            if (colmd != null && colmd.getName() != null)
            {
                name = colmd.getName();
            }
            String oldName = colNameByPosition.put(index, name);
            if (oldName != null)
            {
                throw new NucleusUserException("Error assigning positions of columns in ODF." +
                    " Position=" + index + " was previously assigned to column with name=" + oldName +
                    " but now is for name=" + name +
                    " Check your metadata specification of positions");
            }
        }
    }

    protected static void getColumnInformationForEmbeddedClass(Map<Integer, String> colNameByPosition, 
            AbstractClassMetaData cmd, AbstractMemberMetaData ownerMmd, ClassLoaderResolver clr, MetaDataManager mmgr)
    {
        EmbeddedMetaData emd = ownerMmd.getEmbeddedMetaData();
        AbstractMemberMetaData[] emb_mmd = emd.getMemberMetaData();
        for (int i=0;i<emb_mmd.length;i++)
        {
            AbstractMemberMetaData mmd = cmd.getMetaDataForMember(emb_mmd[i].getName());
            if (emb_mmd[i].getEmbeddedMetaData() == null)
            {
                int index = ODFUtils.getColumnPositionForFieldOfEmbeddedClass(mmd.getAbsoluteFieldNumber(), ownerMmd);
                String name = emb_mmd[i].getName();
                ColumnMetaData[] colmds = emb_mmd[i].getColumnMetaData();
                if (colmds != null && colmds.length > 0 && colmds[0].getName() != null)
                {
                    name = colmds[0].getName();
                }
                String oldName = colNameByPosition.put(index, name);
                if (oldName != null)
                {
                    throw new NucleusUserException("Error assigning positions of columns in ODF." +
                        " Position=" + index + " was previously assigned to column with name=" + oldName +
                        " but now is for name=" + name +
                        " Check your metadata specification of positions");
                }
            }
            else
            {
                int relationType = mmd.getRelationType(clr);
                if (Relation.isRelationSingleValued(relationType))
                {
                    AbstractClassMetaData relCmd = mmgr.getMetaDataForClass(emb_mmd[i].getType(), clr);
                    getColumnInformationForEmbeddedClass(colNameByPosition, relCmd, emb_mmd[i], clr, mmgr);
                }
            }
        }
    }
}