/**********************************************************************
Copyright (c) 2009 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.datanucleus.store.odf.fieldmanager;

import java.lang.reflect.Array;
import java.util.Calendar;
import java.util.Collection;
import java.util.Currency;
import java.util.Date;
import java.util.Map;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.exceptions.NucleusDataStoreException;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.identity.IdentityUtils;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.ColumnMetaData;
import org.datanucleus.metadata.FieldRole;
import org.datanucleus.metadata.MetaDataUtils;
import org.datanucleus.metadata.Relation;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.ObjectProvider;
import org.datanucleus.store.types.sco.SCOUtils;
import org.datanucleus.store.fieldmanager.AbstractFieldManager;
import org.datanucleus.store.odf.ODFUtils;
import org.datanucleus.store.types.ObjectLongConverter;
import org.datanucleus.store.types.ObjectStringConverter;
import org.datanucleus.util.Base64;
import org.datanucleus.util.NucleusLogger;
import org.odftoolkit.odfdom.doc.table.OdfTableCell;
import org.odftoolkit.odfdom.doc.table.OdfTableRow;

/**
 * FieldManager for the fetch of fields from ODF.
 */
public class FetchFieldManager extends AbstractFieldManager
{
    protected final ObjectProvider op;
    ExecutionContext ec;
    AbstractClassMetaData cmd;
    protected final OdfTableRow row;

    public FetchFieldManager(ObjectProvider op, OdfTableRow row)
    {
        this.op = op;
        this.ec = op.getExecutionContext();
        this.cmd = op.getClassMetaData();
        this.row = row;
    }

    public FetchFieldManager(ExecutionContext ec, AbstractClassMetaData cmd, OdfTableRow row)
    {
        this.op = null;
        this.ec = ec;
        this.cmd = cmd;
        this.row = row;
    }

    protected int getColumnIndexForMember(int memberNumber)
    {
        return ODFUtils.getColumnPositionForFieldOfClass(cmd, memberNumber);
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchBooleanField(int)
     */
    @Override
    public boolean fetchBooleanField(int fieldNumber)
    {
        int index = getColumnIndexForMember(fieldNumber);
        OdfTableCell cell = row.getCellByIndex(index);
        return cell.getBooleanValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchByteField(int)
     */
    @Override
    public byte fetchByteField(int fieldNumber)
    {
        int index = getColumnIndexForMember(fieldNumber);
        OdfTableCell cell = row.getCellByIndex(index);
        Double val = cell.getDoubleValue();
        if (val == null)
        {
            return 0;
        }
        return cell.getDoubleValue().byteValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchCharField(int)
     */
    @Override
    public char fetchCharField(int fieldNumber)
    {
        int index = getColumnIndexForMember(fieldNumber);
        OdfTableCell cell = row.getCellByIndex(index);
        return cell.getStringValue().charAt(0);
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchDoubleField(int)
     */
    @Override
    public double fetchDoubleField(int fieldNumber)
    {
        int index = getColumnIndexForMember(fieldNumber);
        OdfTableCell cell = row.getCellByIndex(index);
        Double val = cell.getDoubleValue();
        if (val == null)
        {
            return 0.0;
        }
        return val.doubleValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchFloatField(int)
     */
    @Override
    public float fetchFloatField(int fieldNumber)
    {
        int index = getColumnIndexForMember(fieldNumber);
        OdfTableCell cell = row.getCellByIndex(index);
        Double val = cell.getDoubleValue();
        if (val == null)
        {
            return 0.0f;
        }
        return val.floatValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchIntField(int)
     */
    @Override
    public int fetchIntField(int fieldNumber)
    {
        int index = getColumnIndexForMember(fieldNumber);
        OdfTableCell cell = row.getCellByIndex(index);
        Double val = cell.getDoubleValue();
        if (val == null)
        {
            return 0;
        }
        return val.intValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchLongField(int)
     */
    @Override
    public long fetchLongField(int fieldNumber)
    {
        int index = getColumnIndexForMember(fieldNumber);
        OdfTableCell cell = row.getCellByIndex(index);
        Double val = cell.getDoubleValue();
        if (val == null)
        {
            return 0;
        }
        return val.longValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchShortField(int)
     */
    @Override
    public short fetchShortField(int fieldNumber)
    {
        int index = getColumnIndexForMember(fieldNumber);
        OdfTableCell cell = row.getCellByIndex(index);
        Double val = cell.getDoubleValue();
        if (val == null)
        {
            return 0;
        }
        return val.shortValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchStringField(int)
     */
    @Override
    public String fetchStringField(int fieldNumber)
    {
        int index = getColumnIndexForMember(fieldNumber);
        OdfTableCell cell = row.getCellByIndex(index);
        return cell.getStringValue();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#fetchObjectField(int)
     */
    @Override
    public Object fetchObjectField(int fieldNumber)
    {
        ClassLoaderResolver clr = ec.getClassLoaderResolver();
        AbstractMemberMetaData mmd =
            cmd.getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        int relationType = mmd.getRelationType(clr);

        // Special cases
        if (Relation.isRelationSingleValued(relationType) && mmd.isEmbedded())
        {
            // Persistable object embedded into table of this object
            Class embcls = mmd.getType();
            AbstractClassMetaData embcmd = ec.getMetaDataManager().getMetaDataForClass(embcls, clr);
            if (embcmd != null)
            {
                ObjectProvider embSM = ec.newObjectProviderForEmbedded(mmd, embcmd, op, fieldNumber);
                embSM.replaceFields(embcmd.getAllMemberPositions(), new FetchEmbeddedFieldManager(embSM, row, mmd));
                return embSM.getObject();
            }
        }

        return fetchObjectFieldFromCell(fieldNumber, mmd, clr);
    }

    protected Object fetchObjectFieldFromCell(int fieldNumber, AbstractMemberMetaData mmd, ClassLoaderResolver clr)
    {
        int relationType =  mmd.getRelationType(clr);
        int index = getColumnIndexForMember(fieldNumber);
        OdfTableCell cell = row.getCellByIndex(index);
        if (relationType == Relation.NONE)
        {
            Class type = mmd.getType();
            Object value = null;
            if (type == Boolean.class)
            {
                if (cell.getBooleanValue() == null)
                {
                    return null;
                }
                value = cell.getBooleanValue();
            }
            else if (type == Byte.class)
            {
                if (cell.getDoubleValue() == null)
                {
                    return null;
                }
                value = cell.getDoubleValue().byteValue();
            }
            else if (type == Character.class)
            {
                if (cell.getStringValue() == null)
                {
                    return null;
                }
                value = cell.getStringValue().charAt(0);
            }
            else if (type == Double.class)
            {
                if (cell.getDoubleValue() == null)
                {
                    return null;
                }
                value = cell.getDoubleValue();
            }
            else if (type == Float.class)
            {
                if (cell.getDoubleValue() == null)
                {
                    return null;
                }
                value = new Float(cell.getDoubleValue().floatValue());
            }
            else if (type == Integer.class)
            {
                if (cell.getDoubleValue() == null)
                {
                    return null;
                }
                value = new Integer(cell.getDoubleValue().intValue());
            }
            else if (type == Long.class)
            {
                if (cell.getDoubleValue() == null)
                {
                    return null;
                }
                value = new Long(cell.getDoubleValue().longValue());
            }
            else if (type == Short.class)
            {
                if (cell.getDoubleValue() == null)
                {
                    return null;
                }
                value = new Short(cell.getDoubleValue().shortValue());
            }
            else if (type == Calendar.class)
            {
                if (cell.getDateValue() == null)
                {
                    return null;
                }
                value = cell.getDateValue();
            }
            else if (Date.class.isAssignableFrom(type))
            {
                if (cell.getDateValue() == null)
                {
                    return null;
                }
                Calendar cal = cell.getDateValue();
                value = cal.getTime();
            }
            else if (type == Currency.class)
            {
                if (cell.getStringValue() == null)
                {
                    return null;
                }
                ObjectStringConverter strConv = ec.getNucleusContext().getTypeManager().getStringConverter(mmd.getType());
                value = strConv.toObject(cell.getStringValue());
            }
            else if (Enum.class.isAssignableFrom(type))
            {
                String cellValue = cell.getStringValue();
                if (cellValue != null && cellValue.length() > 0)
                {
                    value = Enum.valueOf(type, cell.getStringValue());
                }
                else
                {
                    return null;
                }
            }
            else if (byte[].class == type)
            {
                String cellValue = cell.getStringValue();
                if (cellValue != null && cellValue.length() > 0)
                {
                    value = Base64.decode(cellValue);
                }
            }
            else
            {
                boolean useLong = false;
                ColumnMetaData[] colmds = mmd.getColumnMetaData();
                if (colmds != null && colmds.length == 1)
                {
                    String jdbc = colmds[0].getJdbcType();
                    if (jdbc != null && (jdbc.equalsIgnoreCase("int") || jdbc.equalsIgnoreCase("integer")))
                    {
                        useLong = true;
                    }
                }
                ObjectStringConverter strConv = ec.getNucleusContext().getTypeManager().getStringConverter(mmd.getType());
                ObjectLongConverter longConv = ec.getNucleusContext().getTypeManager().getLongConverter(mmd.getType());
                if (useLong && longConv != null)
                {
                    value = longConv.toObject(cell.getDoubleValue().longValue());
                }
                else if (!useLong && strConv != null)
                {
                    String cellValue = cell.getStringValue();
                    if (cellValue != null && cellValue.length() > 0)
                    {
                        value = strConv.toObject(cell.getStringValue());
                    }
                    else
                    {
                        return null;
                    }
                }
                else if (!useLong && longConv != null)
                {
                    value = longConv.toObject(cell.getDoubleValue().longValue());
                }
                else
                {
                    // Not supported as String so just set to null
                    NucleusLogger.PERSISTENCE.warn("Field " + mmd.getFullFieldName() + 
                        " could not be set in the object since it is not persistable to ODF");
                    return null;
                }
            }
            if (op != null)
            {
                return op.wrapSCOField(fieldNumber, value, false, false, true);
            }
            return value;
        }
        else if (Relation.isRelationSingleValued(relationType))
        {
            // Persistable object - retrieve the string form of the identity, and find the object
            String idStr = cell.getStringValue();
            if (idStr == null)
            {
                return null;
            }

            if (idStr.startsWith("[") && idStr.endsWith("]"))
            {
                idStr = idStr.substring(1, idStr.length()-1);
                return IdentityUtils.getObjectFromIdString(idStr, mmd, FieldRole.ROLE_FIELD, ec, true);
            }
            else
            {
                return null;
            }
        }
        else if (Relation.isRelationMultiValued(relationType))
        {
            // Collection/Map/Array
            String cellStr = cell.getStringValue();
            if (cellStr == null)
            {
                return null;
            }

            if (cellStr.startsWith("[") && cellStr.endsWith("]"))
            {
                cellStr = cellStr.substring(1, cellStr.length()-1);
                String[] components = MetaDataUtils.getInstance().getValuesForCommaSeparatedAttribute(cellStr);
                if (Collection.class.isAssignableFrom(mmd.getType()))
                {
                    Collection<Object> coll;
                    try
                    {
                        Class instanceType = SCOUtils.getContainerInstanceType(mmd.getType(), mmd.getOrderMetaData() != null);
                        coll = (Collection<Object>) instanceType.newInstance();
                    }
                    catch (Exception e)
                    {
                        throw new NucleusDataStoreException(e.getMessage(), e);
                    }

                    if (components != null)
                    {
                        for (int i=0;i<components.length;i++)
                        {
                            // TODO handle Collection<interface>
                            AbstractClassMetaData elementCmd = mmd.getCollection().getElementClassMetaData(
                                ec.getClassLoaderResolver(), ec.getMetaDataManager());
                            Object element = IdentityUtils.getObjectFromIdString(components[i], elementCmd, ec, true);
                            coll.add(element);
                        }
                    }
                    if (op != null)
                    {
                        return op.wrapSCOField(fieldNumber, coll, false, false, true);
                    }
                    return coll;
                }
                else if (Map.class.isAssignableFrom(mmd.getType()))
                {
                    AbstractClassMetaData keyCmd = mmd.getMap().getKeyClassMetaData(clr, ec.getMetaDataManager());
                    AbstractClassMetaData valCmd = mmd.getMap().getValueClassMetaData(clr, ec.getMetaDataManager());

                    Map map;
                    try
                    {
                        Class instanceType = SCOUtils.getContainerInstanceType(mmd.getType(), false);
                        map = (Map) instanceType.newInstance();
                    }
                    catch (Exception e)
                    {
                        throw new NucleusDataStoreException(e.getMessage(), e);
                    }

                    if (components != null)
                    {
                        for (int i=0;i<components.length;i++)
                        {
                            String keyCmpt = components[i];
                            i++;
                            String valCmpt = components[i];

                            // Strip square brackets from entry bounds
                            String keyStr = keyCmpt.substring(1, keyCmpt.length()-1);
                            String valStr = valCmpt.substring(1, valCmpt.length()-1);

                            Object key = null;
                            if (keyCmd != null)
                            {
                                // TODO handle Map<interface, ?>
                                key = IdentityUtils.getObjectFromIdString(keyStr, keyCmd, ec, true);
                            }
                            else
                            {
                                String keyTypeName = mmd.getMap().getKeyType();
                                Class keyType = ec.getClassLoaderResolver().classForName(keyTypeName);
                                if (Enum.class.isAssignableFrom(keyType))
                                {
                                    key = Enum.valueOf(keyType, keyStr);
                                }
                                else if (keyType == String.class)
                                {
                                    key = keyStr;
                                }
                                else
                                {
                                    // TODO Support other map key types
                                    throw new NucleusException("Don't currently support retrieval of Maps with keys of type " + keyTypeName + " (field="+mmd.getFullFieldName() + ")");
                                }
                            }

                            Object val = null;
                            if (valCmd != null)
                            {
                                // TODO handle Map<?, interface>
                                val = IdentityUtils.getObjectFromIdString(valStr, valCmd, ec, true);
                            }
                            else
                            {
                                String valTypeName = mmd.getMap().getValueType();
                                Class valType = ec.getClassLoaderResolver().classForName(valTypeName);
                                if (Enum.class.isAssignableFrom(valType))
                                {
                                    val = Enum.valueOf(valType, valStr);
                                }
                                else if (valType == String.class)
                                {
                                    val = valStr;
                                }
                                else
                                {
                                    // TODO Support other map value types
                                    throw new NucleusException("Don't currently support retrieval of Maps with values of type " + valTypeName + " (field="+mmd.getFullFieldName() + ")");
                                }
                            }

                            map.put(key, val);
                        }
                    }
                    if (op != null)
                    {
                        return op.wrapSCOField(fieldNumber, map, false, false, true);
                    }
                    return map;
                }
                else if (mmd.getType().isArray())
                {
                    Object array = null;
                    if (components != null)
                    {
                        array = Array.newInstance(mmd.getType().getComponentType(), components.length);
                        for (int i=0;i<components.length;i++)
                        {
                            // TODO handle interface[]
                            AbstractClassMetaData elementCmd = mmd.getCollection().getElementClassMetaData(
                                ec.getClassLoaderResolver(), ec.getMetaDataManager());
                            Object element = IdentityUtils.getObjectFromIdString(components[i], elementCmd, ec, true);
                            Array.set(array, i, element);
                        }
                    }
                    else
                    {
                        array = Array.newInstance(mmd.getType().getComponentType(), 0);
                    }
                    if (op != null)
                    {
                        return op.wrapSCOField(fieldNumber, array, false, false, true);
                    }
                    return array;
                }
            }
            throw new NucleusException("Dont currently support retrieval of collection/map/array types from ODF");
        }
        else
        {
            throw new NucleusException("Dont currently support retrieval of type " + mmd.getTypeName() + " from ODF");
        }
    }
}